import itertools
import logging
import multiprocessing
import os
from urllib.parse import urljoin

import bs4
import urllib3


def get_page_soup(url: str) -> bs4.BeautifulSoup:
    """Get page."""
    conn = urllib3.connection_from_url(url)
    r = conn.request("GET", url)
    data = r.data.decode("utf-8")
    return bs4.BeautifulSoup(data)


def get_page_links(soup: bs4.BeautifulSoup) -> list[str]:
    """Get links in page."""
    return [el.attrs["href"] for el in soup.find_all("a")]


def get_all_file_urls(url: str) -> list[str]:
    """Get all file links by DFS."""
    soup = get_page_soup(url)
    links = get_page_links(soup)
    files = [urljoin(url, a) for a in links if not a.endswith("/")]
    dirs = [urljoin(url, a) for a in links if a != "../" and a.endswith("/")]
    files.extend(itertools.chain(*[get_all_file_urls(dir) for dir in dirs]))
    return files


def download_file(dist_dir: str, url: str) -> None:
    conn = urllib3.connection_from_url(url)
    r = conn.request("GET", url)
    name = url.split("/")[-1]
    if not os.path.exists(dist_dir):
        os.makedirs(dist_dir)

    with open(os.path.join(dist_dir, name), "wb") as file:
        file.write(r.data)


def download_all(
    index_url: str, dist: str, override: bool = False, multiprocess: int = 4
) -> None:
    """Download directory from index page.

    E.g: `https://download.brainimagelibrary.org/biccn/zeng/luo/fMOST/cells/`

    Parameters
    ----------
    index_url : str
        URL of index page.
    dist : str
        Path of target directory.
    override : bool, default `False`
        Override existing file, skip file if `False`.
    multiprocess : int, default `4`
        How many process are available for download.
    """

    all_file_urls = get_all_file_urls(index_url)
    logging.info(
        "downloader: search `{}`, found {} files.", index_url, len(all_file_urls)
    )

    def download(dist: str, url: str) -> None:
        file_dist = url.removeprefix(index_url)
        components = file_dist.split("/")[:-1]
        dir = os.path.join(dist, *components)
        if os.path.exists(file_dist):
            if not override:
                return

            logging.info("downloader: file `{}` exits, deleted.", dir)
            os.remove(file_dist)

        try:
            logging.info("downloader: downloading `{}` to `{}`", url, dir)
            download_file(dir, url)

            logging.info("downloader: download `{}` to `{}`", url, dir)
        except Exception as ex:
            logging.info("downloader: fails to download `{}`, except `{}`", url, ex)

    with multiprocessing.Pool(multiprocess) as p:
        p.map(lambda url: download(dist, url), all_file_urls)



from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.collections import LineCollection


@dataclass
class Palette:
    momo: str = "#F596AA"
    mizugaki: str = "#B9887D"


palette = Palette()


def draw_lines(
    lines: npt.NDArray[np.floating], ax: plt.Axes | None = None, **kwargs
) -> tuple[plt.Axes, LineCollection]:
    """Draw lines.

    Parameters
    ----------
    lines : A collection of coords of lines
        Excepting a ndarray of shape (N, 2, 3), the axis-2 holds two points,
        and the axis-3 holds the coordinates (x, y, z).
    ax : ~matplotlib.axes.Axes, optional
        A subplot. If `None`, a new one will be created.
    **kwargs : dict[str, Unknown]
        Forwarded to `~matplotlib.collections.LineCollection`.

    Returns
    -------
    ax : ~matplotlib.axes.Axes
        If provided, return as-is.
    collection : ~matplotlib.collections.LineCollection
        Drawn line collection.
    """

    starts = lines[:, 0]
    ends = lines[:, 1]

    # transform
    # TODO: support camare view
    starts = starts[:, 0:2]
    ends = ends[:, 0:2]

    edges = np.stack((starts, ends), axis=1)
    collection = LineCollection(edges, **kwargs)  # type: ignore

    if ax is None:
        fig, ax = plt.subplots(1, 1)
    ax.add_collection(collection)
    ax.set_aspect(1)
    ax.autoscale()
    ax.axis("off")
    if ax is None:
        plt.show()

    return ax, collection



from ....core import BranchTree, Tree
from ...transforms import Transform


class ToBranchTree(Transform[Tree, BranchTree]):
    """Transform tree to branch tree."""

    def __init__(self) -> None:
        super().__init__("branchtree")

    def apply(self, x: Tree) -> BranchTree:
        return BranchTree.from_tree(x)



import itertools
from typing import Callable, Optional, overload

from typing_extensions import Self  # TODO: move to typing in python 3.11

from .branch import Branch
from .tree import K, T, Tree


class BranchTree(Tree):
    """A branch tree that contains only soma, branch, and tip nodes."""

    class Node(Tree.Node):
        # fmt:off
        @property
        def branches(self) -> list[Branch]: return self["branches"]
        @branches.setter
        def branches(self, v: list[Branch]): self["branches"] = v
        # fmt:on

        def __init__(
            self,
            id: int,
            type: int,
            x: float,
            y: float,
            z: float,
            r: float,
            pid: int,
            **kwargs,
        ) -> None:
            kwargs.setdefault("branches", [])
            super().__init__(id, type, x, y, z, r, pid, **kwargs)

    TraverseEnter = Callable[[Node, Optional[T]], T]
    TraverseLeave = Callable[[Node, list[T]], T]

    def get_branches(self) -> list[Branch]:
        return self.traverse(leave=lambda n, p: list(itertools.chain(n.branches, *p)))

    # fmt:off
    @overload
    def traverse(self, *, enter: TraverseEnter[T]) -> None: ...
    @overload
    def traverse(self, *, enter: Optional[TraverseEnter[T]] = None, leave: TraverseLeave[K]) -> K: ...
    # fmt:on

    def traverse(
        self,
        *,
        enter: Optional[TraverseEnter[T]] = None,
        leave: Optional[TraverseLeave[K]] = None,
    ) -> K | None:
        return super().traverse(enter=enter, leave=leave)  # type: ignore

    def __getitem__(self, idx: int) -> Node:
        return super().__getitem__(idx)  # type: ignore

    @classmethod
    def from_swc(cls, swc_path: str) -> Self:
        """Generating a branch tree from swc file.

        Parameters
        ----------
        swc_path : str
            Path of *.swc.

        Returns
        -------
        BranchTree
            A branch tree.
        """

        tree = super().from_swc(swc_path)
        return cls.from_tree(tree)

    @classmethod
    def from_tree(cls, tree: Tree) -> Self:
        """Generating a branch tree from tree.

        Parameters
        ----------
        tree : Tree
            A neuron tree.

        Returns
        -------
        BranchTree : BranchTree
            A branch tree.
        """

        self = cls()
        self._source = tree._source

        def reducer(old_id: int, parent_id: Optional[int]) -> list[Tree.Node]:
            node = cls.Node(**tree[old_id])  # make shallow copy
            neighbors = list(tree.G.neighbors(old_id))
            if (parent_id is not None) and (len(neighbors) == 1):
                branch_nodes = reducer(neighbors[0], parent_id)
                branch_nodes.append(node)
                return branch_nodes

            node.id = len(self) + 1
            node.pid = parent_id if parent_id is not None else -1
            self._add_node(node)
            if parent_id is not None:
                self._add_edge(parent_id, node.id)

            for n in neighbors:
                branch_nodes = reducer(n, node.id)
                branch_nodes.append(node)
                branch_nodes.reverse()
                node.branches.append(Branch(branch_nodes))

            return [node]

        reducer(self.root, None)
        return self



import functools
import os
import random
from typing import Any, Callable, Iterator, Optional, TypeVar, cast, overload

import matplotlib.collections
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import numpy.typing as npt
import pandas as pd
from typing_extensions import Self  # TODO: move to typing in python 3.11

from ..utils import painter

T, K = TypeVar("T"), TypeVar("K")


class Tree:
    """A neuron tree, which should be a binary tree in most cases."""

    class Node(dict[str, Any]):
        """Node of neuron tree"""

        def __init__(
            self,
            id: int,
            type: int,
            x: float,
            y: float,
            z: float,
            r: float,
            pid: int,
            **kwargs,
        ) -> None:
            super().__init__(id=id, type=type, x=x, y=y, z=z, r=r, pid=pid, **kwargs)

        # fmt: off
        @property
        def id(self) -> int: return self["id"]
        @id.setter
        def id(self, v: int): self["id"] = v
        @property
        def type(self) -> int: return self["type"]
        @type.setter
        def type(self, v: int): self["type"] = v
        @property
        def x(self) -> float: return self["x"]
        @x.setter
        def x(self, v: float): self["x"] = v
        @property
        def y(self) -> float: return self["y"]
        @y.setter
        def y(self, v: float): self["y"] = v
        @property
        def z(self) -> float: return self["z"]
        @z.setter
        def z(self, v: float): self["z"] = v
        @property
        def r(self) -> float: return self["r"]
        @r.setter
        def r(self, v: float): self["r"] = v
        @property
        def pid(self) -> int: return self["pid"]
        @pid.setter
        def pid(self, v: int): self["pid"] = v
        # fmt: on

        def xyz(self) -> npt.NDArray[np.float64]:
            """Get the `x`, `y`, `z` of branch, an array of shape (3,)"""
            return np.array([self.x, self.y, self.z], dtype=np.float64)

        def xyzr(self) -> npt.NDArray[np.float64]:
            """Get the `x`, `y`, `z`, `r` of branch, an array of shape (4,)"""
            return np.array([self.x, self.y, self.z, self.r], dtype=np.float64)

        def distance(self, b: "Tree.Node") -> float:
            """Get the distance of two nodes."""
            return np.linalg.norm(self.xyz() - b.xyz()).item()

        def format_swc(self) -> str:
            """Get the SWC format string."""
            x, y, z, r = ["%.4f" % f for f in [self.x, self.y, self.z, self.r]]
            items = [self.id, self.type, x, y, z, r, self.pid]
            return " ".join(map(str, items))

        def __str__(self) -> str:
            return self.format_swc()

        @classmethod
        def from_dataframe_row(cls, row: tuple[Any, ...]) -> Self:
            """Read node from row of `~pandas.DataFrame`"""
            keys = ("id", "type", "x", "y", "z", "r", "pid")
            return cls(*[getattr(row, key) for key in keys])

    root: int
    G: nx.DiGraph
    nodes: dict[int, Node]
    scale: tuple[float, float, float, float]
    _source: Optional[str] = None

    def __init__(self) -> None:
        self.G = nx.DiGraph()
        self.root = 1  # default to 1
        self.scale = (1, 1, 1, 1)
        self.nodes = {}

    def to_swc(self, swc_path: str) -> None:
        """Write swc file."""
        with open(swc_path, "w") as f:
            if self._source is not None:
                f.write(f"# source: {self._source}\n")

            f.write("# id type x y z r pid\n")
            f.writelines((str(n) + "\n" for n in self))

    def copy(self, G: bool = True) -> Self:
        """Make a copy.

        Parameters
        ----------
        G : bool, default `True`
            Skip copy G if false.
        """
        newTree = Tree()
        newTree._source = self._source
        newTree.G = self.G.copy() if G else self.G
        newTree.nodes = {k: v for k, v in self.nodes.items()} if G else self.nodes
        return newTree

    def draw(
        self,
        color: Optional[str] = painter.palette.momo,
        ax: Optional[plt.Axes] = None,
        **kwargs,
    ) -> tuple[plt.Axes, matplotlib.collections.LineCollection]:
        """Draw neuron tree.

        Parameters
        ----------
        color : str, optional
            Color of branch. If `None`, the default color will be enabled.
        ax : ~matplotlib.axes.Axes, optional
            A subplot of `~matplotlib`. If `None`, a new one will be created.
        **kwargs : dict[str, Unknown]
            Forwarded to `~matplotlib.collections.LineCollection`.

        Returns
        -------
        ax : ~matplotlib.axes.Axes
            If provided, return as-is.
        collection : ~matplotlib.collections.LineCollection
            Drawn line collection.
        """
        edges = np.array([[self[a].xyz(), self[b].xyz()] for a, b in self.G.edges])
        return painter.draw_lines(edges, ax=ax, color=color, **kwargs)

    TraverseEnter = Callable[[Node, Optional[T]], T]
    TraverseLeave = Callable[[Node, list[T]], T]

    # fmt:off
    @overload
    def traverse(self, *, enter: TraverseEnter[T]) -> None: ...
    @overload
    def traverse(self, *, enter: Optional[TraverseEnter[T]] = None, leave: TraverseLeave[K]) -> K: ...
    # fmt:on

    def traverse(
        self,
        *,
        enter: Optional[TraverseEnter[T]] = None,
        leave: Optional[TraverseLeave[K]] = None,
    ) -> K | None:
        """Traverse each nodes.

        Parameters
        ----------
        enter : Callable[[Node, list[T]], T], optional
            The callback when entering each node, it accepts two parameters,
            the first parameter is the current node, the second parameter is
            the parent's information T, and the root node receives an None.
        leave : Callable[[Node, Optional[T]], T], optional
            The callback when leaving each node. When leaving a node, subtree
            has already been traversed. Callback accepts two parameters, the
            first parameter is the current node, the second parameter is the
            children's information T, and the leaf node receives an empty list.
        """
        return self._traverse(enter, leave, self.root, None)

    def length(self) -> float:
        """Sum of length of all segments."""
        return self.traverse(
            leave=lambda n, acc: sum(acc)
            + sum(map(lambda b: n.distance(self[b]), self.G.neighbors(n.id)))
        )

    def random_cut(self, keep_percent: float) -> Self:
        """Random cut tree.

        Parameters
        ----------
        keep_percent : float
            The percent of preserved segment length.

        Returns
        -------
        Tree
            A new tree.
        """
        tree = self.copy()
        length = (1 - keep_percent) * self.length()
        while tree.G.size() > 1 and length > 0:
            nodes = [i for i, d in tree.G.out_degree() if d == 0]
            i = nodes[random.randint(0, len(nodes) - 1)]
            predecessors = list(tree.G.predecessors(i))
            if len(predecessors) == 0:
                continue
            length -= tree[predecessors[0]].distance(tree[i])
            tree.G.remove_node(i)

        return tree

    def normalize(self) -> None:
        """Normalize neuron tree.

        Scale the `x`, `y`, `z`, `r` of nodes to 0-1
        """

        _min, _max = self.traverse(
            leave=lambda a, children: functools.reduce(
                lambda acc, cur: (
                    np.min(np.stack([acc[0], cur[0]]).transpose(), axis=1),
                    np.max(np.stack([acc[1], cur[1]]).transpose(), axis=1),
                ),
                children,
                (a.xyzr(), a.xyzr()),
            )
        )
        scale = 1 / (_max - _min)
        self.scale = tuple(scale)

        def scaler(a: Tree.Node, _):
            a.x, a.y, a.z, a.r = (a.xyzr() - _min) * scale

        self.traverse(leave=scaler)

    def standardize(self) -> None:
        raise NotImplementedError()

    def _add_edge(self, id: int, child_id: int) -> None:
        self.G.add_edge(id, child_id)

    def _add_node(self, node: Node) -> None:
        self.G.add_node(node.id)
        self.nodes[node.id] = node

    def _traverse(
        self,
        enter: Optional[TraverseEnter[T]],
        leave: Optional[TraverseLeave[K]],
        idx: int,
        pre: Optional[T],
    ) -> K | None:
        cur = enter(self[idx], pre) if enter is not None else None
        children = [self._traverse(enter, leave, i, cur) for i in self.G.neighbors(idx)]
        return leave(self[idx], cast(list[K], children)) if leave is not None else None

    def __getitem__(self, idx: int) -> Node:
        """Get node by id."""
        return self.nodes[idx]

    def __iter__(self) -> Iterator[Node]:
        """Iter each nodes."""
        return (self[i] for i in self.G)

    def __len__(self) -> int:
        """Get number of nodes."""
        return len(self.G)

    def __str__(self) -> str:
        nodes, edges = self.G.number_of_nodes(), self.G.number_of_edges()
        return f"Neuron Tree with {nodes} nodes and {edges} edges"

    @classmethod
    def from_swc(cls, swc_path: str, names: Optional[list[str]] = None) -> Self:
        """Read neuron tree from swc file."""

        self = cls()
        self._source = os.path.abspath(swc_path)

        names = ["id", "type", "x", "y", "z", "r", "pid"] if names is None else names
        df = pd.read_csv(
            swc_path,
            sep=" ",
            comment="#",
            names=names,
            dtype={
                "id": np.int16,
                "type": np.int8,
                "x": np.float64,
                "y": np.float64,
                "z": np.float64,
                "r": np.float64,
                "pid": np.int16,
            },
        )

        nodes = [cls.Node.from_dataframe_row(r) for r in df.itertuples()]
        self.G.add_nodes_from([n.id for n in nodes])
        self.G.add_edges_from([(n.pid, n.id) for n in nodes if n.pid != -1])
        self.nodes = {n.id: n for n in nodes}
        self.root = df.iloc[0]["id"]
        return self



from swcgeom.core.tree import Tree
from swcgeom.data.torch.transforms.tree_transforms import ToBranchTree
from swcgeom.data.transforms import Transforms

from ... import BranchTree
from .tree_folder_dataset import TreeFolderDataset


# TODO: remove in 0.2.0
class BranchTreeFolderDataset(TreeFolderDataset):
    """Branch tree folder in swc format.

    .. deprecated:: 0.1.7
        `BranchTreeFolderDataset` will be removed in swcgeom 0.2.0, it is
        replaced by `TreeFolderDataset(self.swc_dir, transforms=transforms)`
        because the latter provides a more flexible approach.
    """

    def __init__(self, swc_dir: str) -> None:
        """Create tree dataset.

        Parameters
        ----------
        swc_dir : str
            Path of SWC file directory.
        """
        transforms = Transforms[Tree, BranchTree](ToBranchTree())
        super().__init__(swc_dir, transforms=transforms)

    def __getitem__(self, idx: int) -> tuple[BranchTree, int]:
        """Get a branch tree.

        Returns
        -------
        x : BranchTree
            A branch tree from swc format.
        y : int
            Label of x.
        """
        x, y = super().__getitem__(idx)
        return BranchTree.from_tree(x), y



from typing import Any, Generic, Iterable, TypeVar, cast, overload

T, K = TypeVar("T"), TypeVar("K")

T1, T2, T3 = TypeVar("T1"), TypeVar("T2"), TypeVar("T3")
T4, T5, T6 = TypeVar("T4"), TypeVar("T5"), TypeVar("T6")


class Transform(Generic[T, K]):
    """An abstract class representing a :class:`Transform`.

    All transforms that represent a map from `T` to `K`. All subclasses should
    overwrite :meth:`apply`, supporting applying transform in `x`. Subclasses
    could also optionally overwrite :meth:`apply_batch`, which is expected to
    applying transform a batch of input, manual implementation usually gives
    better performance.
    """

    name: str

    def __init__(self, name: str) -> None:
        """"""
        super().__init__()
        self.name = name

    def __call__(self, x: T) -> K:
        """Apply transform."""
        return self.apply(x)

    def apply(self, x: T) -> K:
        """Apply transform."""
        raise NotImplementedError()

    def apply_batch(self, batch: Iterable[T]) -> Iterable[K]:
        """Apply transform in batch."""
        return map(self.apply, batch)

    def get_name(self) -> str:
        """Get transform name."""
        return self.name


class Transforms(list[Transform[Any, Any]], Generic[T, K]):
    """A simple typed wrapper for transforms."""

    # fmt:off
    @overload
    def __init__(self, t1: Transform[T, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, T5], t6: Transform[T5, K], /) -> None: ...
    @overload
    def __init__(self, t1: Transform[T, T1],  t2: Transform[T1, T2],
                       t3: Transform[T2, T3], t4: Transform[T3, T4],
                       t5: Transform[T4, T5], t6: Transform[T5, T6],
                       t7: Transform[T6, Any], /, *transforms: Transform[Any, Any]) -> None: ...
    # fmt:on

    def __init__(self, *transforms: Transform[Any, Any]) -> None:
        super().__init__(transforms)

    def __call__(self, x: T) -> K:
        """Apply transforms."""
        return self.apply(x)

    def apply(self, x: T) -> K:
        """Apply transforms."""
        for transform in self:
            x = transform.apply(x)

        return cast(K, x)

    def apply_batch(self, batch: Iterable[T]) -> Iterable[K]:
        """Apply transforms in batch."""
        for transform in self:
            batch = transform.apply_batch(batch)

        return cast(Iterable[K], batch)

    def get_name(self, separator: str = "_") -> str:
        """Get name of transforms.

        Parameters
        ----------
        separator : str, default `_`

        Returns
        -------
        name : str
            Name which join separator to the names.
        """
        return separator.join(self.get_names())

    def get_names(self) -> Iterable[str]:
        """Get names of transforms."""
        return (transform.get_name() for transform in self)



import os
from typing import Generic, TypeVar, cast

from swcgeom.data.transforms import Transforms

import torch.utils.data

from ... import Tree

T = TypeVar("T")


class TreeFolderDataset(torch.utils.data.Dataset, Generic[T]):
    """Tree folder in swc format."""

    swc_dir: str
    swcs: list[str]

    transforms: Transforms[Tree, T] | None

    def __init__(
        self,
        swc_dir: str,
        transforms: Transforms[Tree, T] | None = None,
    ) -> None:
        """Create tree dataset.

        Parameters
        ----------
        swc_dir : str
            Path of SWC file directory.
        transfroms : Transfroms[Tree, T], optional
            Branch transformations.

        See Also
        --------
        ~swcgeom.data.torch.transforms : module
            Preset transform set.
        """
        super().__init__()
        self.swc_dir = swc_dir
        self.swcs = self.find_swcs(swc_dir)
        self.transforms = transforms

    def __getitem__(self, idx: int) -> tuple[T, int]:
        """Get a tree data.

        Returns
        -------
        x : Tree
            A tree from swc format.
        y : int
            Label of x, always 0.
        """
        tree = Tree.from_swc(self.swcs[idx])
        x = self.transforms.apply(tree) if self.transforms else tree
        return cast(T, x), 0

    def __len__(self) -> int:
        """Get length of set of trees."""
        return len(self.swcs)

    @staticmethod
    def find_swcs(swc_dir: str) -> list[str]:
        """Find all swc files."""
        swcs = list[str]()
        for root, dirs, files in os.walk(swc_dir):
            files = [f for f in files if os.path.splitext(f)[-1] == ".swc"]
            swcs.extend([os.path.join(root, f) for f in files])

        return swcs



from typing import Callable, Literal

import torch

from ....core import Branch
from ...transforms import Transform


class BranchResampler(Transform[Branch, Branch]):
    """Resample branch."""

    def __init__(self, num: int) -> None:
        super().__init__(f"resample{self.num}")
        self.num = num

    def apply(self, x: Branch) -> Branch:
        return x.resample("linear", num=self.num)


class BranchStandardize(Transform[Branch, Branch]):
    """Standarize branch."""

    def __init__(self) -> None:
        super().__init__("standardized")

    def apply(self, x: Branch) -> Branch:
        return x.standardize()


class BranchToTensor(Transform[Branch, torch.Tensor]):
    """Transform branch to ~torch.Tensor."""

    channels: str
    channel_first: bool

    fn: Callable[[Branch], torch.Tensor]

    def __init__(
        self, channels: Literal["xyz", "xyzr"] = "xyz", channel_first: bool = True
    ):
        """Transform branch to ~torch.Tensor.

        Parameters
        ----------
        channels : str, default `xyz`
            Output channels and order, support `xyz` and `xyzr` now.
        channel_first : bool
            Tensor dimension order. If `True`, the tensor is of shape (C, N), otherwise of
            shape (N, C).
        """

        super().__init__("tensor")
        self.channels = channels
        self.channel_first = channel_first

        match self.channels:
            case "xyz":
                self.fn = lambda x: torch.from_numpy(x.xyz()).float()
            case "xyzr":
                self.fn = lambda x: torch.from_numpy(x.xyzr()).float()
            case _:
                raise ValueError("invalid channel.")

    def apply(self, x: Branch) -> torch.Tensor:
        output = self.fn(x)
        return output.T if self.channel_first else output



import numpy as np
import numpy.typing as npt


def angle(a: npt.ArrayLike, b: npt.ArrayLike) -> float:
    """Get the agnle of vectors.

    Returns
    -------
    angle : float
        Angle in radians.
    """
    a, b = np.array(a), np.array(b)
    costheta = np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))
    theta = np.arccos(costheta)
    return theta if np.cross(a, b) > 0 else -theta


def scale3d(sx: float, sy: float, sz: float) -> npt.NDArray[np.float64]:
    """Get the 3D scale transfomation matrix.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [sx, 0, 0, 0],
            [0, sy, 0, 0],
            [0, 0, sz, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float64,
    )


def translate3d(tx: float, ty: float, tz: float) -> npt.NDArray[np.float64]:
    """Get the 3D translate transfomation matrix.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [1, 0, 0, tx],
            [0, 1, 0, ty],
            [0, 0, 1, tz],
            [0, 0, 0, 1],
        ],
        dtype=np.float64,
    )


def rotate3d(n: npt.ArrayLike, theta: float) -> npt.NDArray[np.float64]:
    r"""Get the 3D rotation transfomation matrix.
    
    Rotate v with axis n by an angle theta according to the right hand rule, 
    follow rodrigues' rotaion formula.

    .. math::

        R(\mathbf{n}, \alpha) = \cos{\alpha} \cdot \mathbf{I}
        + (1-\cos{\alpha}) \cdot \mathbf{n} \cdot \mathbf{n^T}
        + \sin{\alpha} \cdot \mathbf{N} \\

        N  = \begin{pmatrix}
        0 & -n_z & n_y \\
        n_z & 0 & -n_x \\
        -n_y & n_x & 0
        \end{pmatrix}

    Parameters
    ----------
    n : ArrayLike
        Rotation axis.
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """

    n = np.array(n)
    nx, ny, nz = n[0:3]
    N = np.array(
        [
            [0, -nz, ny],
            [nz, 0, -nx],
            [-ny, nx, 0],
        ],
        dtype=np.float64,
    )

    return (
        np.cos(theta) * np.identity(4)
        + (1 - np.cos(theta)) * n * n[:, None]
        + np.sin(theta) * N
    )


def rotate3d_x(theta: float) -> npt.NDArray[np.float64]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `x`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """

    return np.array(
        [
            [1, 0, 0, 0],
            [0, np.cos(theta), -np.sin(theta), 0],
            [0, np.sin(theta), np.cos(theta), 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float64,
    )


def rotate3d_y(theta: float) -> npt.NDArray[np.float64]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `y`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [np.cos(theta), 0, np.sin(theta), 0],
            [0, 1, 0, 0],
            [-np.sin(theta), 0, np.cos(theta), 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float64,
    )


def rotate3d_z(theta: float) -> npt.NDArray[np.float64]:
    """Get the 3D rotation transfomation matrix.

    Rotate 3D vector `v` with `z`-axis by an angle theta according to the right
    hand rule.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.

    Returns
    -------
    T : np.NDArray
        The homogeneous transfomation matrix, shape (4, 4).
    """
    return np.array(
        [
            [np.cos(theta), -np.sin(theta), 0, 0],
            [np.sin(theta), np.cos(theta), 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ],
        dtype=np.float64,
    )



from typing import Iterable, Literal, NamedTuple, overload

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.collections import LineCollection
from typing_extensions import Self  # TODO: move to typing in python 3.11

from ..utils import painter, transforms
from .tree import Tree

Scale = NamedTuple("Scale", x=float, y=float, z=float, r=float)


class Branch(list[Tree.Node]):
    """A branch of neuron tree.

    Notes
    -----
    Only a part of data of branch nodes is valid, such as `x`, `y`, `z` and
    `r`, but the `id` and `type` is usually invalid.
    """

    scale: Scale

    def __init__(self, nodes: Iterable[Tree.Node]) -> None:
        super().__init__(nodes)
        self.scale = Scale(1, 1, 1, 1)

    def x(self) -> npt.NDArray[np.float64]:
        """Get the `x` of branch.

        Returns
        -------
        x : np.ndarray[np.float64]
            An array of shape (n_sample,).
        """
        return np.array([n.x for n in self], dtype=np.float64)

    def y(self) -> npt.NDArray[np.float64]:
        """Get the `y` of branch.

        Returns
        -------
        y : np.ndarray[np.float64]
            An array of shape (n_sample,).
        """
        return np.array([n.y for n in self], dtype=np.float64)

    def z(self) -> npt.NDArray[np.float64]:
        """Get the `z` of branch.

        Returns
        -------
        z : np.ndarray[np.float64]
            An array of shape (n_sample,).
        """
        return np.array([n.z for n in self], dtype=np.float64)

    def r(self) -> npt.NDArray[np.float64]:
        """Get the `r` of branch.

        Returns
        -------
        r : np.ndarray[np.float64]
            An array of shape (n_sample,).
        """
        return np.array([n.r for n in self], dtype=np.float64)

    def xyz(self) -> npt.NDArray[np.float64]:
        """Get the `x`, `y`, `z` of branch.

        Returns
        -------
        xyz : np.ndarray[np.float64]
            An array of shape (n_sample, 3).
        """
        return np.array([n.xyz() for n in self], dtype=np.float64)

    def xyzr(self) -> npt.NDArray[np.float64]:
        """Get the `x`, `y`, `z`, `r` of branch.

        Returns
        -------
        xyzr : np.ndarray[np.float64]
            An array of shape (n_sample, 4).
        """
        return np.array([n.xyzr() for n in self], dtype=np.float64)

    def draw(
        self,
        color: str = painter.palette.mizugaki,
        ax: plt.Axes | None = None,
        standardize: bool = True,
        **kwargs,
    ) -> tuple[plt.Axes, LineCollection]:
        """Draw neuron branch.

        Parameters
        ----------
        color : str, optional
            Color of branch. If `None`, the default color will be enabled.
        ax : ~matplotlib.axes.Axes, optional
            A subplot. If `None`, a new one will be created.
        standardize : bool, default `True`
            Standardize branch, see also self.standardize.
        **kwargs : dict[str, Any]
            Forwarded to `matplotlib.collections.LineCollection`.

        Returns
        -------
        ax : ~matplotlib.axes.Axes
            If provided, return as-is.
        collection : ~matplotlib.collections.LineCollection
            Drawn line collection.
        """

        if standardize:
            xyzr, _ = self._standardize()
            xyz = xyzr[:, :3]
        else:
            xyz = self.xyz()

        lines = np.array([xyz[:-1], xyz[1:]]).swapaxes(0, 1)
        return painter.draw_lines(lines, color=color, ax=ax, **kwargs)

    # fmt:off
    @overload
    def resample(self, *, num: int) -> Self: ...
    @overload
    def resample(self, mode: Literal["linear"], *, num: int) -> Self: ...
    # fmt:on

    def resample(self, mode: str = "linear", **kwargs) -> Self:
        """Resample branch to special num of nodes.

        Parameters
        ----------
        mode : str, default `linear`
            Resample mode.

            Linear mode :
                Resampling by linear interpolation, DO NOT keep original node.

                num : int
                    Number of nodes after resample.

        See Also
        --------
        cls.resample_linear : linear mode.
        """

        xyzr = self.xyzr()
        if mode == "linear":
            new_xyzr = self.resample_linear(xyzr, **kwargs)
        else:
            raise Exception(f"unsupported reample mode '{mode}'")

        return Branch.from_numpy(new_xyzr)

    def standardize(self) -> Self:
        """Standarize a branch.

        Standardized branch starts at (0, 0, 0), ends at (1, 0, 0), up at y,
        and scale max radius to 1.
        """

        new_xyzr, scale = self._standardize()
        branch = Branch.from_numpy(new_xyzr)
        branch.scale = scale
        return branch

    def length(self) -> float:
        """Sum of length of stems."""
        xyz = self.xyz()
        return np.sum(np.linalg.norm(xyz[1:] - xyz[:-1], axis=1))

    def straight_line_distance(self) -> float:
        """Distance between start point and end point."""
        return float(np.linalg.norm(self[-1].xyz() - self[0].xyz()))

    def _standardize(self) -> tuple[npt.NDArray[np.float64], Scale]:
        xyzr = self.xyzr()
        xyz, r = xyzr[:, 0:3], xyzr[:, 3:4]
        T, s = self.get_standardize_matrix(xyzr)

        ones = np.ones([xyz.shape[0], 1])
        xyz4 = np.concatenate([xyz, ones], axis=1).transpose()  # (4, N)
        new_xyz = np.dot(T, xyz4)[0:3, :].transpose()
        new_xyzr = np.concatenate([new_xyz, r * s.r], axis=1)

        return new_xyzr, Scale(*np.multiply(s, self.scale))

    def __str__(self) -> str:
        return f"Neuron branch with {len(self)} nodes."

    @classmethod
    def from_numpy(cls, xyzr: npt.NDArray[np.float64]) -> Self:
        """Create a branch from ~numpy.ndarray.

        Parameters
        ----------
        xyzr : npt.NDArray[np.float64]
            Collection of nodes. If shape (n, 4), both `x`, `y`, `z`, `r` of
            nodes is enabled. If shape (n, 3), only `x`, `y`, `z` is enabled
            and `r` will fill by 1.
        """
        assert xyzr.ndim == 2
        assert xyzr.shape[1] >= 3

        if xyzr.shape[1] == 3:
            ones = np.ones([xyzr.shape[0], 1])
            xyzr = np.concatenate([xyzr, ones], axis=1)

        nodes = [
            Tree.Node(i, 0, x, y, z, r, i - 1)
            for i, (x, y, z, r) in zip(range(xyzr.shape[0]), xyzr[:, 0:4])
        ]
        return cls(nodes)

    @classmethod
    def from_numpy_batch(cls, xyzr_batch: npt.NDArray[np.float64]) -> list[Self]:
        """Create list of branch form ~numpy.ndarray.

        Parameters
        ----------
        xyzr : npt.NDArray[np.float64]
            Batch of collection of nodes. If shape (bs, n, 4), both `x`, `y`,
            `z`, `r` of nodes is enabled. If shape (bs, n, 3), only `x`, `y`,
            `z` is enabled and `r` will fill by 1.
        """

        assert xyzr_batch.ndim == 3
        assert xyzr_batch.shape[1] >= 3

        if xyzr_batch.shape[2] == 3:
            ones = np.ones([xyzr_batch.shape[0], xyzr_batch.shape[1], 1])
            xyzr_batch = np.concatenate([xyzr_batch, ones], axis=2)

        branches = list[Branch]()
        for xyzr in xyzr_batch:
            nodes = [
                Tree.Node(i, 0, x, y, z, r, i - 1)
                for i, (x, y, z, r) in zip(range(xyzr.shape[0]), xyzr[:, 0:4])
            ]
            branches.append(Branch(nodes))

        return branches

    @staticmethod
    def resample_linear(
        xyzr: npt.NDArray[np.float64], num: int
    ) -> npt.NDArray[np.float64]:
        """Resampling by linear interpolation, DO NOT keep original node.

        Parameters
        ----------
        xyzr : np.ndarray[np.float64]
            The array of shape (N, 4).
        num : int
            Number of nodes after resample.

        Returns
        -------
        coords : ~numpy.NDArray[float64]
            An array of shape (num, 4).
        """

        xp = np.cumsum(np.linalg.norm(xyzr[1:, :3] - xyzr[:-1, :3], axis=1))
        xp = np.insert(xp, 0, 0)
        xvals = np.linspace(0, xp[-1], num)

        x = np.interp(xvals, xp, xyzr[:, 0])
        y = np.interp(xvals, xp, xyzr[:, 1])
        z = np.interp(xvals, xp, xyzr[:, 2])
        r = np.interp(xvals, xp, xyzr[:, 3])
        return np.stack([x, y, z, r], axis=1)

    @staticmethod
    def get_standardize_matrix(
        xyzr: npt.NDArray[np.float64],
    ) -> tuple[npt.NDArray[np.float64], Scale]:
        """Get standarize transformation matrix.

        Standardized branch starts at (0, 0, 0), ends at (1, 0, 0), up at y,
        and scale max radius to 1.

        Parameters
        ----------
        xyzr : np.ndarray[np.float64]
            The `x`, `y`, `z`, `r` matrix of shape (N, 4) of branch.

        Returns
        -------
        T : np.ndarray[np.float64]
            An homogeneous transfomation matrix of shape (4, 4).
        scale : float
            Scale ratio.
        """
        assert xyzr.ndim == 2
        assert xyzr.shape[1] == 4

        xyz, r = xyzr[:, :3], xyzr[:, 3:]
        T = np.identity(4)
        v = np.concatenate([xyz[-1] - xyz[0], np.zeros((1))])[:, None]

        # translate to the origin
        translate = transforms.translate3d(-xyz[0, 0], -xyz[0, 1], -xyz[0, 2])
        T = np.dot(translate, T)

        # scale to unit vector
        s = float(1 / np.linalg.norm(v[:3, 0]))
        scale = transforms.scale3d(s, s, s)
        T = np.dot(scale, T)

        # rotate v to the xz-plane, v should be (x, 0, z) now
        vy = np.dot(T, v)[:, 0]
        # when looking at the xz-plane along the positive y-axis, the
        # coordinates should be (z, x)
        rotate_y = transforms.rotate3d_y(transforms.angle([vy[2], vy[0]], [0, 1]))
        T = np.dot(rotate_y, T)

        # rotate v to the x-axis, v should be (1, 0, 0) now
        vx = np.dot(T, v)[:, 0]
        rotate_z = transforms.rotate3d_z(transforms.angle([vx[0], vx[1]], [1, 0]))
        T = np.dot(rotate_z, T)

        # rotate the farthest point to the xy-plane
        if xyz.shape[0] > 2:
            ones = np.ones([xyz.shape[0], 1])
            xyz4 = np.concatenate([xyz, ones], axis=1).transpose()  # (4, N)
            new_xyz4 = np.dot(T, xyz4)  # (4, N)
            max_index = np.argmax(np.linalg.norm(new_xyz4[1:3, :], axis=0)[1:-1]) + 1
            max_xyz4 = xyz4[:, max_index].reshape(4, 1)
            max_xyz4_t = np.dot(T, max_xyz4)  # (4, 1)
            angle_x = transforms.angle(max_xyz4_t[1:3, 0], [1, 0])
            rotate_x = transforms.rotate3d_x(angle_x)
            T = np.dot(rotate_x, T)

        # scale max radius to 1
        sr = 1 / r.max()

        return T, Scale(s, s, s, sr)



import os
import warnings
from typing import Generic, Iterable, Optional, TypeVar, cast

import numpy as np

import torch
import torch.utils.data

from ...core import Branch
from .transforms import ToBranchTree, Transforms
from .tree_folder_dataset import TreeFolderDataset

T = TypeVar("T")


class BranchDataset(torch.utils.data.Dataset, Generic[T]):
    """An easy way to get branches."""

    swc_dir: str
    save: str | None
    transforms: Transforms[Branch, T] | None

    branches: list[T]

    def __init__(
        self,
        swc_dir: str,
        save: str | bool = True,
        transforms: Optional[Transforms[Branch, T]] = None,
    ) -> None:
        """Create branch dataset.

        Parameters
        ----------
        swc_dir : str
            Path of SWC file directory.
        save : Union[str, bool], default `True`
            Save branch data to file if not False. If `True`, automatically
            generate file name.
        transfroms : Transfroms[Branch, T], optional
            Branch transformations.

        See Also
        --------
        ~swcgeom.data.torch.transforms : module
            Preset transform set.
        """

        self.swc_dir = swc_dir
        self.transforms = transforms

        if isinstance(save, str):
            self.save = save
        elif save:
            self.save = os.path.join(swc_dir, self.get_filename())
        else:
            self.save = None

        if self.save and os.path.exists(self.save):
            self.branches = torch.load(self.save)
            return

        self.branches = self.get_branches()
        if self.save:
            torch.save(self.branches, self.save)

    def __getitem__(self, idx: int) -> tuple[T, int]:
        """Get branch data.

        Returns
        -------
        x : T
            Transformed data.
        y : int
            Label of x.
        """
        return self.branches[idx], 0

    def __len__(self) -> int:
        """Get length of branches."""
        return len(self.branches)

    def get_filename(self) -> str:
        """Get filename."""
        names = self.transforms.get_names() if self.transforms else []
        name = "_".join(["branch_dataset", *names])
        return f"{name}.pt"

    def get_branches(self) -> list[T]:
        """Get all branches."""
        transforms = Transforms(ToBranchTree())
        branch_trees = TreeFolderDataset(self.swc_dir, transforms=transforms)
        branches = list[T]()
        old_settings = np.seterr(all="raise")
        for x, y in branch_trees:
            try:
                brs = x.get_branches()
                brs = self.transforms.apply_batch(brs) if self.transforms else brs
                branches.extend(cast(Iterable[T], brs))
            except Exception as ex:
                warnings.warn(
                    f"BranchDataset: skip swc '{x}', got warning from numpy: {ex}"
                )

        np.seterr(**old_settings)
        return branches
